import min from"./min.js";import rotate from"./rotate.js";import verifyOptimum from"./verifyOptimum.js";import checkDelta2 from"./checkDelta2.js";import checkDelta3 from"./checkDelta3.js";import statistics from"./statistics.js";import endpoints from"./endpoints.js";import neighbours from"./neighbours.js";import blossomLeaves from"./blossomLeaves.js";import blossomEdges from"./blossomEdges.js";// Adapted from http://jorisvr.nl/maximummatching.html
// All credit for the implementation goes to Joris van Rantwijk [http://jorisvr.nl].
// ** Original introduction below **
// Weighted maximum matching in general graphs.
// The algorithm is taken from "Efficient Algorithms for Finding Maximum
// Matching in Graphs" by Zvi Galil, ACM Computing Surveys, 1986.
// It is based on the "blossom" method for finding augmenting paths and
// the "primal-dual" method for finding a matching of maximum weight, both
// due to Jack Edmonds.
// Some ideas came from "Implementation of algorithms for maximum matching
// on non-bipartite graphs" by H.J. Gabow, Standford Ph.D. thesis, 1973.
// A C program for maximum weight matching by Ed Rothberg was used extensively
// to validate this new code.
export default function blossom(a,b){b===void 0&&(b=!1),a===void 0&&(a=!0);/**
	 * Compute a maximum-weighted matching in the general undirected
	 * weighted graph given by "edges". If "maxCardinality" is true,
	 * only maximum-cardinality matchings are considered as solutions.
	 *
	 * Edges is a sequence of tuples (i, j, wt) describing an undirected
	 * edge between vertex i and vertex j with weight wt.  There is at most
	 * one edge between any two vertices; no vertex has an edge to itthis.
	 * Vertices are identified by consecutive, non-negative integers.
	 *
	 * Return a list "mate", such that mate[i] === j if vertex i is
	 * matched to vertex j, and mate[i] === -1 if vertex i is not matched.
	 *
	 * This function takes time O(n^3)
	 *
	 * @param {Array} edges
	 * @param {Boolean} maxCardinality
	 * @return {Array}
	 */const c=(c,e=!1)=>{// Vertices are numbered 0 .. (nvertex-1).
// Non-trivial blossoms are numbered nvertex .. (2*nvertex-1)
//
// Edges are numbered 0 .. (nedge-1).
// Edge endpoints are numbered 0 .. (2*nedge-1), such that endpoints
// (2*k) and (2*k+1) both belong to edge k.
//
// Many terms used in the comments (sub-blossom, T-vertex) come from
// the paper by Galil; read the paper before reading this code.
// Deal swiftly with empty graphs.
if(0===c.length)return[];// Count vertices + find the maximum edge weight.
const[f,g,h]=statistics(c),l=endpoints(g,c),m=neighbours(f,g,c),n=Array(f).fill(-1),o=Array(2*f).fill(0),q=Array(2*f).fill(-1),r=Array(f);// If p is an edge endpoint,
// endpoint[p] is the vertex to which endpoint p is attached.
// Not modified by the algorithm.
for(let a=0;a<f;++a)r[a]=a;// If b is a sub-blossom,
// blossomparent[b] is its immediate parent (sub-)blossom.
// If b is a top-level blossom, blossomparent[b] is -1.
const u=Array(2*f).fill(-1),x=Array(2*f).fill(null),y=Array(2*f);// If b is a non-trivial (sub-)blossom,
// blossomchilds[b] is an ordered list of its sub-blossoms, starting with
// the base and going round the blossom.
for(let a=0;a<f;++a)y[a]=a;y.fill(-1,f,2*f);// If b is a non-trivial (sub-)blossom,
// blossomendps[b] is a list of endpoints on its connecting edges,
// such that blossomendps[b][i] is the local endpoint of blossomchilds[b][i]
// on the edge that connects it to blossomchilds[b][wrap(i+1)].
const z=Array(2*f).fill(null),A=Array(2*f).fill(-1),B=Array(2*f).fill(null),C=Array(f);// If v is a free vertex (or an unreached vertex inside a T-blossom),
// bestedge[v] is the edge to an S-vertex with least slack,
// or -1 if there is no such edge.
// If b is a (possibly trivial) top-level S-blossom,
// bestedge[b] is the least-slack edge to a different S-blossom,
// or -1 if there is no such edge.
// This is used for efficient computation of delta2 and delta3.
for(let a=0;a<f;++a)C[a]=f+a;// If v is a vertex,
// dualvar[v] = 2 * u(v) where u(v) is the v's variable in the dual
// optimization problem (multiplication by two ensures integer values
// throughout the algorithm if all edge weights are integers).
// If b is a non-trivial blossom,
// dualvar[b] = z(b) where z(b) is b's variable in the dual optimization
// problem.
const D=Array(2*f);D.fill(h,0,f),D.fill(0,f,2*f);// If allowedge[k] is true, edge k has zero slack in the optimization
// problem; if allowedge[k] is false, the edge's slack may or may not
// be zero.
const E=Array(g).fill(!1);// Queue of newly discovered S-vertices.
let F=[];// Return 2 * slack of edge k (does not work inside blossoms).
const G=a=>{const[b,d,e]=c[a];return D[b]+D[d]-2*e},H=(a,c,d)=>{const e=r[a];if(o[a]=c,o[e]=c,q[a]=d,q[e]=d,A[a]=-1,A[e]=-1,1===c)// B became an S-vertex/blossom; add it(s vertices) to the queue.
for(const a of blossomLeaves(f,x,e))F.push(a);else{// B became a T-vertex/blossom; assign label S to its mate.
// (If b is a non-trivial blossom, its base is the only vertex
// with an external mate.)
const a=y[e];H(l[n[a]],1,1^n[a])}},I=(a,c)=>{// Trace back from v and w, placing breadcrumbs as we go.
const d=[];let e=-1;for(;-1!==a||-1!==c;){// Look for a breadcrumb in v's blossom or put a new breadcrumb.
let f=r[a];if(4&o[f]){e=y[f];break}// Swap v and w so that we alternate between both paths.
if(d.push(f),o[f]=5,-1===q[f]?a=-1:(a=l[q[f]],f=r[a],a=l[q[f]]),-1!==c){const b=a;a=c,c=b}}// Remove breadcrumbs.
for(const e of d)o[e]=1;// Return base vertex, if we found one.
return e},J=(a,d)=>{let e=c[d][0],g=c[d][1];const h=r[a];let i=r[e],j=r[g];// Create blossom.
const n=C.pop();y[n]=a,u[n]=-1,u[h]=n;// Make list of sub-blossoms and their interconnecting edge endpoints.
const b=[];x[n]=b;const p=[];// Trace back from v to base.
for(z[n]=p;i!==h;)// Add bv to the new blossom.
u[i]=n,b.push(i),p.push(q[i]),e=l[q[i]],i=r[e];// Reverse lists, add endpoint that connects the pair of S vertices.
// Trace back from w to base.
for(b.push(h),b.reverse(),p.reverse(),p.push(2*d);j!==h;)// Add bw to the new blossom.
u[j]=n,b.push(j),p.push(1^q[j]),g=l[q[j]],j=r[g];// Set label to S.
o[n]=1,q[n]=q[h],D[n]=0;// Relabel vertices.
for(const b of blossomLeaves(f,x,n))2===o[r[b]]&&F.push(b),r[b]=n;// Compute blossombestedges[b].
const s=Array(2*f).fill(-1),t=b.length;for(let e=0;e<t;++e){const a=b[e];// Walk this subblossom's least-slack edges.
let d=B[a];null===d&&(d=blossomEdges(f,x,m,a));for(const a of d){const[b,d]=c[a],e=r[d]===n?r[b]:r[d];e!==n&&1===o[e]&&(-1===s[e]||G(a)<G(s[e]))&&(s[e]=a)}// Forget about least-slack edges of the subblossom.
B[a]=null,A[a]=-1}B[n]=[];const E=s.length;for(let b=0;b<E;++b)d=s[b],-1!==d&&B[n].push(d);// Select bestedge[b].
const H=B[n].length;if(0<H){A[n]=B[n][0];for(let a=1;a<H;++a)d=B[n][a],G(d)<G(A[n])&&(A[n]=d)}else A[n]=-1},K=(a,b)=>{// Convert sub-blossoms into top-level blossoms.
for(let c=0;c<x[a].length;++c){const d=x[a][c];if(u[d]=-1,d<f)r[d]=d;else if(b&&0===D[d])K(d,b);else for(const a of blossomLeaves(f,x,d))r[a]=d}// If we expand a T-blossom during a stage, its sub-blossoms must be
// relabeled.
if(!b&&2===o[a]){// Start at the sub-blossom through which the expanding
// blossom obtained its label, and relabel sub-blossoms untili
// we reach the base.
// Figure out through which sub-blossom the expanding blossom
// obtained its label initially.
const b=r[l[1^q[a]]];// Decide in which direction we will go round the blossom.
let c,d,e,g,h=x[a].indexOf(b);1&h?(c=1,d=0,e=x[a].length,g=0):(c=-1,d=1,e=0,g=x[a].length);// Move along the blossom until we get to the base.
let i=q[a];for(;h!==e;)// Relabel the T-sub-blossom.
// Step to the next S-sub-blossom and note its forward endpoint.
// Step to the next T-sub-blossom.
o[l[1^i]]=0,o[l[1^(z[a][h-d]^d)]]=0,H(l[1^i],2,i),E[Math.floor(z[a][h-d]/2)]=!0,h+=c,i=z[a][h-d]^d,E[Math.floor(i/2)]=!0,h+=c;// Relabel the base T-sub-blossom WITHOUT stepping through to
// its mate (so don't call assignLabel).
let j=x[a][0];for(o[l[1^i]]=2,o[j]=2,q[l[1^i]]=i,q[j]=i,A[j]=-1,h=g+c;x[a][h]!==b;){if(j=x[a][h],1===o[j]){h+=c;continue}for(const a of blossomLeaves(f,x,j))if(0!==o[a]){o[a]=0,o[l[n[y[j]]]]=0,H(a,2,q[a]);break}// If the sub-blossom contains a reachable vertex, assign
// label T to the sub-blossom.
h+=c}}// Recycle the blossom number.
o[a]=-1,q[a]=-1,x[a]=null,z[a]=null,y[a]=-1,B[a]=null,A[a]=-1,C.push(a)},L=(a,b)=>{// Bubble up through the blossom tree from vertex v to an immediate
// sub-blossom of b.
let c,d,e,g,h,k=b;for(;u[k]!==a;)k=u[k];// Recursively deal with the first sub-blossom.
k>=f&&L(k,b),c=x[a].indexOf(k);const m=c,i=x[a].length;// Move along the blossom until we get to the base.
for(1&m?(d=1,e=0,g=i):(d=-1,e=1,g=0);c!==g;)// Step to the next sub-blossom and augment it recursively.
// Step to the next sub-blossom and augment it recursively.
// Match the edge connecting those sub-blossoms.
c+=d,k=x[a][c],h=z[a][c-e]^e,k>=f&&L(k,l[h]),c+=d,k=x[a][Math.abs(c%i)],k>=f&&L(k,l[1^h]),n[l[h]]=1^h,n[l[1^h]]=h;// Rotate the list of sub-blossoms to put the new base at the front.
rotate(x[a],m),rotate(z[a],m),y[a]=y[x[a][0]]},M=a=>{const b=c[a][0],d=c[a][1];N(b,2*a+1),N(d,2*a)},N=(a,b)=>{// Match vertex s to remote endpoint p. Then trace back from s
// until we find a single vertex, swapping matched and unmatched
// edges as we go.
// eslint-disable-next-line no-constant-condition
for(;;){const c=r[a];// Trace one step back.
if(c>=f&&L(c,a),n[a]=b,-1===q[c])// Reached single vertex; stop.
break;const d=l[q[c]],e=r[d];a=l[q[e]];const g=l[1^q[e]];// Augment through the T-blossom from j to base.
e>=f&&L(e,g),n[g]=q[e],b=1^q[e]}};// Assign label t to the top-level blossom containing vertex w
// and record the fact that w was reached through the edge with
// remote endpoint p.
let O,j,s,t,w,P,Q;// Main loop: continue until no further improvement is possible.
for(let a=0;a<f;++a){o.fill(0),A.fill(-1),B.fill(null,f,2*f),E.fill(!1),F=[];// Label single blossoms/vertices with S and put them in the queue.
for(let a=0;a<f;++a)-1===n[a]&&0===o[r[a]]&&H(a,1,-1);// Loop until we succeed in augmenting the matching.
let a=!1;// eslint-disable-next-line no-constant-condition
for(;;){// Each iteration of this loop is a "substage".
// A substage tries to find an augmenting path;
// if found, the path is used to improve the matching and
// the stage ends. If there is no augmenting path, the
// primal-dual method is used to pump some slack out of
// the dual variables.
// Continue labeling until all vertices which are reachable
// through an alternating path have got a label.
for(;0<F.length&&!a;){// Take an S vertex from the queue.
const c=F.pop(),b=m[c].length;for(let d=0;d<b;++d){const b=m[c][d],e=Math.floor(b/2),f=l[b];// W is a neighbour to v
if(r[c]!==r[f])// This edge is internal to a blossom; ignore it
if(E[e]||(j=G(e),0>=j&&(E[e]=!0)),E[e]){if(0===o[r[f]])H(f,2,1^b);else if(1!==o[r[f]])0===o[f]&&(// W is inside a T-blossom, but w itthis has not
// yet been reached from outside the blossom;
// mark it as reached (we need this to relabel
// during T-blossom expansion).
o[f]=2,q[f]=1^b);else if(s=I(c,f),0<=s)J(s,e);else{M(e),a=!0;break}}else if(1===o[r[f]]){// Keep track of the least-slack non-allowable edge to
// a different S-blossom.
const a=r[c];(-1===A[a]||j<G(A[a]))&&(A[a]=e)}else 0===o[f]&&(-1===A[f]||j<G(A[f]))&&(A[f]=e)}}if(a)break;// There is no augmenting path under these constraints;
// compute delta and reduce slack in the optimization problem.
// (Note that our vertex dual variables, edge slacks and delta's
// are pre-multiplied by two.)
t=-1,w=null,P=null,Q=null,b&&(checkDelta2({nvertex:f,neighbend:m,label:o,endpoint:l,bestedge:A,slack:G,inblossom:r}),checkDelta3({nvertex:f,edges:c,blossomparent:u,blossomchilds:x,neighbend:m,label:o,endpoint:l,bestedge:A,slack:G,inblossom:r})),e||(t=1,w=min(D,0,f));// Compute delta2: the minimum slack on any edge between
// an S-vertex and a free vertex.
for(let a=0;a<f;++a)0===o[r[a]]&&-1!==A[a]&&(O=G(A[a]),(-1===t||O<w)&&(w=O,t=2,P=A[a]));// Compute delta3: half the minimum slack on any edge between
// a pair of S-blossoms.
for(let a=0;a<2*f;++a)-1===u[a]&&1===o[a]&&-1!==A[a]&&(j=G(A[a]),O=j/2,(-1===t||O<w)&&(w=O,t=3,P=A[a]));// Compute delta4: minimum z variable of any T-blossom.
for(let a=f;a<2*f;++a)0<=y[a]&&-1===u[a]&&2===o[a]&&(-1===t||D[a]<w)&&(w=D[a],t=4,Q=a);-1===t&&(t=1,w=Math.max(0,min(D,0,f)));// Update dual variables according to delta.
for(let a=0;a<f;++a)1===o[r[a]]?// S-vertex: 2*u = 2*u - 2*delta
D[a]-=w:2===o[r[a]]&&(// T-vertex: 2*u = 2*u + 2*delta
D[a]+=w);for(let a=f;a<2*f;++a)0<=y[a]&&-1===u[a]&&(1===o[a]?D[a]+=w:2===o[a]&&(D[a]-=w));// Take action at the point where minimum delta occurred.
if(1===t)// No further improvement possible; optimum reached.
break;else if(2===t){E[P]=!0;let a=c[P][0];0===o[r[a]]&&(a=c[P][1]),F.push(a)}else if(3===t){E[P]=!0;const a=c[P][0];F.push(a)}else// Expand the least-z blossom.
K(Q,!1)}// End of a this substage.
// Stop when no more augmenting path can be found.
if(!a)break;// End of a stage; expand all S-blossoms which have dualvar = 0.
for(let a=f;a<2*f;++a)-1===u[a]&&0<=y[a]&&1===o[a]&&0===D[a]&&K(a,!0)}// Verify that we reached the optimum solution.
a&&verifyOptimum({nvertex:f,edges:c,maxCardinality:e,nedge:g,blossomparent:u,mate:n,endpoint:l,dualvar:D,blossombase:y,blossomendps:z});// Transform mate[] such that mate[v] is the vertex to which v is paired.
for(let a=0;a<f;++a)0<=n[a]&&(n[a]=l[n[a]]);for(let a=0;a<f;++a);return n};return c}
//# sourceMappingURL=blossom.js.map